from typing import Optional
import os

from pyneuphonic._voices import Voices
from pyneuphonic._sse import SSEClient, AsyncSSEClient
from pyneuphonic._endpoint import Endpoint
from pyneuphonic._websocket import AsyncWebsocketClient


class Neuphonic:
    """
    The client for Neuphonic's TTS (text-to-speech) python library.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
    ):
        """Constructor for the Neuphonic client.

        Parameters
        ----------
        api_key
            Your API key. Generate this on https://beta.neuphonic.com. If this is not passed in,
            it needs to be set in your environment and retrievable via `os.getenv('NEUPHONIC_API_TOKEN')`
        base_url : Optional[str], optional
            The base url pointing to which regional deployment to use. If this is not passed on
            and not set in `os.getenv('NEUPHONIC_API_URL')`, then it will default to
            'eu-west-1.api.neuphonic.com'.
        """
        if api_key is None:
            api_key = os.getenv('NEUPHONIC_API_TOKEN')

            if api_key is None:
                raise EnvironmentError(
                    '`api_key` has not been passed in and `NEUPHONIC_API_TOKEN` is not set in the environment.'
                )

        if base_url is None:
            base_url = os.getenv('NEUPHONIC_API_URL')

            if base_url is None:
                base_url = 'eu-west-1.api.neuphonic.com'

        self._api_key = api_key
        self._base_url = base_url

        self.voices = Voices(api_key=self._api_key, base_url=self._base_url)
        self.tts = TTS(api_key=self._api_key, base_url=self._base_url)


class TTS(Endpoint):
    def SSEClient(self, timeout: Optional[int] = 10) -> SSEClient:
        return SSEClient(
            api_key=self._api_key, base_url=self._base_url, timeout=timeout
        )

    def AsyncSSEClient(self, timeout: Optional[int] = 10) -> AsyncSSEClient:
        return AsyncSSEClient(
            api_key=self._api_key, base_url=self._base_url, timeout=timeout
        )

    def AsyncWebsocketClient(self) -> AsyncWebsocketClient:
        return AsyncWebsocketClient(api_key=self._api_key, base_url=self._base_url)



from pydantic import BaseModel as BaseModel, field_validator, ConfigDict
from typing import List, Optional, Callable, Awaitable, Union
import base64
from enum import Enum


class TTSConfig(BaseModel):
    """
    See https://docs.neuphonic.com/api-reference#options for all available options
    """

    model_config = ConfigDict(extra='allow')

    speed: Optional[float] = 1.0
    temperature: Optional[float] = 0.5
    model: Optional[str] = 'neu_fast'
    voice: Optional[str] = None  # if None, default is used.
    sampling_rate: Optional[int] = 22050
    encoding: Optional[str] = 'pcm_linear'
    language_id: Optional[str] = 'en'

    def to_query_params(self) -> str:
        """Generate a query params string from the TTSConfig object, dropping None values."""
        params = to_dict(self)
        return '&'.join(f'{key}={value}' for key, value in params.items())


class WebsocketEvents(Enum):
    OPEN: str = 'open'
    MESSAGE: str = 'message'
    CLOSE: str = 'close'
    ERROR: str = 'error'


def to_dict(model: BaseModel):
    """Returns a pydantic model as dict, with all of the None items removed."""
    return {k: v for k, v in model.model_dump().items() if v is not None}


class VoiceItem(BaseModel):
    model_config = ConfigDict(extra='allow')
    model_config['protected_namespaces'] = ()

    id: str
    name: str
    tags: List[str] = []
    model_availability: List[str] = []


class VoicesResponse(BaseModel):
    """Response from /voices endpoint."""

    model_config = ConfigDict(extra='allow')

    class VoicesData(BaseModel):
        voices: List[VoiceItem]

    data: VoicesData


class AudioData(BaseModel):
    """Structure of audio data received when using any client."""

    model_config = ConfigDict(extra='allow')

    audio: bytes
    text: Optional[str] = None
    sampling_rate: Optional[int] = None

    @field_validator('audio', mode='before')
    def validate(cls, v: Union[str, bytes]) -> bytes:
        """Convert the received audio from the server into bytes that can be played."""
        if isinstance(v, str):
            return base64.b64decode(v)
        elif isinstance(v, bytes):
            return v

        raise ValueError('`audio` must be a base64 encoded string or bytes.')


class WebsocketResponse(BaseModel):
    """Structure of responses when using AsyncWebsocketClient"""

    model_config = ConfigDict(extra='allow')

    data: AudioData


class SSEResponse(BaseModel):
    """Structure of response when using SSEClient or AsyncSSEClient."""

    model_config = ConfigDict(extra='allow')

    status_code: int
    data: AudioData


class SSERequest(BaseModel):
    """Structure of request when using SSEClient or AsyncSSEClient."""

    model_config = ConfigDict(extra='allow')

    text: str
    model: TTSConfig


class WebsocketEventHandlers(BaseModel):
    open: Optional[Callable[[], Awaitable[None]]] = None
    message: Optional[Callable[[AudioData], Awaitable[None]]] = None
    close: Optional[Callable[[], Awaitable[None]]] = None
    error: Optional[Callable[[Exception], Awaitable[None]]] = None



import asyncio
import websockets
from typing import Callable, Union
import json

from pyneuphonic._endpoint import Endpoint
from pyneuphonic.models import (
    WebsocketEventHandlers,
    TTSConfig,
    WebsocketResponse,
    WebsocketEvents,
)


class AsyncWebsocketClient(Endpoint):
    def __init__(
        self,
        api_key: str,
        base_url: str,
    ):
        super().__init__(api_key=api_key, base_url=base_url)

        self.event_handlers = WebsocketEventHandlers()
        self.message_queue = asyncio.Queue()

        self._ws = None
        self._tasks = []

    def on(self, event: WebsocketEvents, handler: Callable):
        if event not in WebsocketEvents:
            raise ValueError(f'Event "{event}" is not a valid event.')

        setattr(self.event_handlers, event.value, handler)

    async def open(self, tts_config: Union[TTSConfig, dict] = TTSConfig()):
        url = f'{self.ws_url}/speak/{tts_config.language_id}?{tts_config.to_query_params()}'

        self._ws = await websockets.connect(
            url,
            ssl=self.ssl_context,
            extra_headers=self.headers,
        )

        if self.event_handlers.open is not None:
            await self.event_handlers.open()

        receive_task = asyncio.create_task(self._receive())
        self._tasks.append(receive_task)

    async def _receive(self):
        try:
            async for message in self._ws:
                if isinstance(message, str):
                    message = WebsocketResponse(**json.loads(message))

                    if self.event_handlers.message is not None:
                        await self.event_handlers.message(message)
                    else:
                        await self.message_queue.put(message)
        except Exception as e:
            if self.event_handlers.error is not None:
                await self.event_handlers.error(e)
        finally:
            if self.event_handlers.close:
                await self.event_handlers.close()

            await self.close()

    async def send(self, message: Union[str, dict], autocomplete=False):
        assert isinstance(
            message, (str, dict)
        ), 'Message must be an instance of str or dict'
        message = message if isinstance(message, str) else json.dumps(message)
        await self._ws.send(message)

        if autocomplete:
            await self._ws.send(json.dumps({'text': ' <STOP>'}))

    async def complete(self):
        await self.send({'text': ' <STOP>'}, autocomplete=False)

    async def receive(self):
        return await self.message_queue.get()

    async def close(self):
        for task in self._tasks:
            task.cancel()

            try:
                await task
            except asyncio.CancelledError:
                pass

        await self._ws.close()



import re
import importlib.util
import logging


def split_text(text):
    return re.findall(r'\S+\s*', text)


def import_if_installed(package_name, error_message=None):
    if error_message is None:
        error_message = f'{package_name} package is not installed'

    if importlib.util.find_spec(package_name) is not None:
        try:
            return importlib.import_module(package_name)
        except ImportError as e:
            logging.warning(f'Could not import {package_name}: {e}')
            return None
    else:
        raise ImportError(error_message)



import json
import logging
import ssl
import certifi
import os
import asyncio
import websockets
import importlib.util
import warnings

from pyneuphonic.websocket import (
    setup_pyaudio,
    play_audio,
    teardown_pyaudio,
)
from typing import Optional, Callable, Awaitable
from types import MethodType


class NeuphonicWebsocketClient:
    def __init__(
        self,
        NEUPHONIC_API_TOKEN: str = None,
        NEUPHONIC_API_URL: str = None,
        on_message: Optional[
            Callable[['NeuphonicWebsocketClient', dict], Awaitable[None]]
        ] = None,
        on_open: Optional[
            Callable[['NeuphonicWebsocketClient'], Awaitable[None]]
        ] = None,
        on_close: Optional[
            Callable[['NeuphonicWebsocketClient'], Awaitable[None]]
        ] = None,
        on_error: Optional[
            Callable[['NeuphonicWebsocketClient', Exception], Awaitable[None]]
        ] = None,
        on_send: Optional[
            Callable[['NeuphonicWebsocketClient', str], Awaitable[None]]
        ] = None,
        language_id: str = 'en',
        play_audio: bool = True,
        sampling_rate: int = 22050,
        logger: Optional[logging.Logger] = None,
        timeout: Optional[float] = None,
        params: dict = None,
    ):
        """
        This websocket implementation has been DEPRECATED.

        Websocket client for the Neuphonic TTS Engine.

        This client is initialised with the provided callbacks.
        If `play_audio` is set to `True` and PyAudio is installed, the client will automatically detect PyAudio and use it
        to play audio.
        The callbacks should all take the instance of this class as the first argument, and the type signatures should
        be as per the provided type hints.
        The callbacks are called when the corresponding event occurs.

        Parameters
        ----------
        NEUPHONIC_API_TOKEN
            The API token for the Neuphonic TTS Engine. If this is not provided the it is loaded from the
            environment.
        NEUPHONIC_API_URL
            The URL for the Neuphonic TTS Engine websocket. This should be of the form `{aws-region}.api.neuphonic.com`
            Default is `eu-west-1.api.neuphonic.com`
        on_message
            The callback function to be called when a message is received from the websocket server.
        on_open
            The callback function to be called when the websocket connection is opened.
        on_close
            The callback function to be called when the websocket connection is closed.
        on_error
            The callback function to be called when an error occurs.
        on_send
            The callback function to be called when a message is sent to the websocket server.
        language_id
            The language id for the language to generate audio in. Currently only english ('en') is
            available. Default is 'en'.
        play_audio
            Whether to play audio from the websocket server automatically. This is true by default and will use pyaudio
            to play the audio. This will not affect any other callbacks passed in and will run alongside them.
        sampling_rate
            If play_audio is True, this is the sampling_rate to use for the default pyaudio player. Default
            is 22050 Hz.
        logger
            The logger to be used by the client. If not provided, a logger will be created.
        timeout
            The timeout for the websocket client.
        params
            Additional model parameters to be passed to the websocket connection. These will be added as
            query parameters to the websocket URL.
        """
        if NEUPHONIC_API_TOKEN is None:
            NEUPHONIC_API_TOKEN = os.getenv('NEUPHONIC_API_TOKEN')

            if NEUPHONIC_API_TOKEN is None:
                raise EnvironmentError(
                    'NEUPHONIC_API_TOKEN has not been passed in and is not set in the environment.'
                )

        if NEUPHONIC_API_URL is None:
            NEUPHONIC_API_URL = os.getenv('NEUPHONIC_API_URL')

            if NEUPHONIC_API_URL is None:
                NEUPHONIC_API_URL = 'eu-west-1.api.neuphonic.com'

        if not logger:
            logger = logging.getLogger(__name__)

        self._logger = logger

        self._NEUPHONIC_API_TOKEN = NEUPHONIC_API_TOKEN
        self._NEUPHONIC_API_URL = NEUPHONIC_API_URL
        self._timeout = timeout

        self._ws = None
        self._listen_task = None
        self._last_sent_message = None
        self._last_received_message = None
        self._language_id = language_id
        self._play_audio = play_audio
        self._sampling_rate = sampling_rate

        self._bind_callbacks(
            on_message,
            on_open,
            on_close,
            on_error,
            on_send,
        )

        self.params = params

    def _bind_callbacks(
        self,
        on_message,
        on_open,
        on_close,
        on_error,
        on_send,
    ):
        """
        Initialises the callbacks for the websocket client. Binds callbacks to the class instance.
        """
        self._logger.debug('Initialising callbacks.')

        if on_message:
            self.on_message = MethodType(on_message, self)

        if on_open:
            self.on_open = MethodType(on_open, self)

        if on_close:
            self.on_close = MethodType(on_close, self)

        if on_error:
            self.on_error = MethodType(on_error, self)

        if on_send:
            self.on_send = MethodType(on_send, self)

        if importlib.util.find_spec('pyaudio') is not None and self._play_audio:
            self._logger.debug('Using PyAudio to play audio.')
            # if pyaudio is installed, and no callbacks are provided, use the pyaudio callbacks to play audio
            self.setup_pyaudio = MethodType(setup_pyaudio, self)
            self.teardown_pyaudio = MethodType(teardown_pyaudio, self)
            self.play_audio = MethodType(play_audio, self)
        elif importlib.util.find_spec('pyaudio') is None and self._play_audio:
            warnings.warn(
                'Parameter `play_audio` is True but PyAudio is not installed. No audio will be played.',
                UserWarning,
            )
            self._play_audio = False
        else:
            self._play_audio = False

        self._logger.debug('Callbacks initialised.')

    async def _create_ws_connection(self, ping_interval, ping_timeout):
        """
        Creates the websocket connection and saves it into self._ws

        This function is called by the open function and should not be called directly.

        Parameters
        ----------
        ping_interval : int
            The number of seconds to wait between every PING.
        ping_timeout : int
            The number of seconds to wait for a PONG from the websocket server before assuming a timeout error.
        """
        self._logger.debug(
            f'Creating connection with WebSocket Server: {self._NEUPHONIC_API_URL}, params: {self.params}',
        )

        # Construct the URL with query parameters
        protocol = 'wss' if 'localhost' not in self._NEUPHONIC_API_URL else 'ws'
        url = f'{protocol}://{self._NEUPHONIC_API_URL}/speak/{self._language_id}'
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        self._logger.debug('Creating Encrypted SLL Connection.')

        if self.params:
            query_string = '&'.join([f'{k}={v}' for k, v in self.params.items()])
            url += f'?{query_string}'

        self._ws = await websockets.connect(
            url,
            ssl=ssl_context if protocol == 'wss' else None,
            timeout=self._timeout,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            extra_headers={'x-api-key': self._NEUPHONIC_API_TOKEN},
        )

        self._logger.debug(
            f'WebSocket connection has been established: {self._NEUPHONIC_API_URL}',
        )

    async def send(self, message: str, autocomplete=False):
        """
        Send a string to the Neuphonic websocket.

        Parameters
        ----------
        message : str
            The string to send to the websocket.
        autocomplete : bool
            Whether or not to autocomplete the message.
        """
        if self._ws and message:
            self._logger.debug(
                f'Sending message to Neuphonic WebSocket Server: {message}'
            )
            self._last_sent_message = message

            if isinstance(message, dict):
                await self._ws.send(json.dumps(message))
            else:
                await self._ws.send(message)

            await self.on_send(message)

            if autocomplete:
                await self.complete()

        else:
            self._logger.debug('Failed to send message, no WebSocket Server available')

    async def complete(self):
        """Function to complete generation of a segment of audio.

        This function sends the completion token '<STOP>' to the server. This gets the server to
        generate audio up to the end of all the text sent over so far.
        """
        await self.send({'text': ' <STOP>'}, autocomplete=False)

    async def ping(self):
        try:
            await self._ws.ping()
            self._logger.debug('Ping sent to WebSocket server.')
        except Exception as e:
            self._logger.error(f'Error sending ping: {e}')
            await self.on_error(e)

    async def _handle_message(self):
        """
        Handle incoming messages from the websocket server.

        This function is called by the listen function and should not be called directly.
        """
        async for message in self._ws:
            if isinstance(message, str):
                message = json.loads(message)

            self._last_received_message = message
            await self.on_message(message)

            if self._play_audio:
                await self.play_audio(message)

    async def open(self, ping_interval: int = 20, ping_timeout: int = None):
        """
        Open the websocket connection and start listening to incoming messages.

        Parameters
        ----------
        ping_interval : int
            The number of seconds to wait between every PING.
        ping_timeout : int
            The number of seconds to wait for a PONG from the websocket server before assuming a timeout error.
        """
        try:
            await self._create_ws_connection(ping_interval, ping_timeout)
            await self.on_open()
            await self._listen()

            if self._play_audio:
                await self.setup_pyaudio(sampling_rate=self._sampling_rate)
        except Exception as e:
            self._logger.error(f'Error opening WebSocket connection: {e}')
            await self.on_error(e)

    async def _listen(self):
        """
        Start listening to the server and handling responses.

        This function is called by the open function and should not be called directly.
        """

        async def _listen_task(client):
            if client._ws.open:  # if the client is open
                try:
                    exception = None
                    receive_task = asyncio.create_task(client._handle_message())
                    await receive_task
                except Exception as e:
                    client._logger.debug(f'Error in WebSocket process: {e}')
                    exception = e
                finally:
                    await client.close()

                    if exception:
                        await self.on_error(exception)

        self._listen_task = asyncio.create_task(_listen_task(self))

    async def close(self):
        """
        Close the websocket connection and call the NeuphonicWebsocketClient.on_close function.
        """

        async def cancel_listen_task():
            """Stop listening to the websocket responses"""
            try:
                self._listen_task.cancel()
                await self._listen_task
            except asyncio.CancelledError as e:
                pass

        if self._listen_task:
            await cancel_listen_task()

        if self._ws and self._ws.open:
            await self._ws.close()
            await self.on_close()

        if self._play_audio:
            await self.teardown_pyaudio()

        self._logger.debug('Websocket connection closed.')

    async def on_message(self, message: dict):
        """
        This function is called on every incoming message. It receives the message as a python dict object.

        Parameters
        ----------
        message : dict
            The incoming message from the websocket server.
        """
        pass

    async def on_open(self):
        """Called after the websocket connection opens."""
        pass

    async def on_close(self):
        """Called after the websocket connection closes."""
        pass

    async def on_error(self, e: Exception):
        """
        Called on error.

        Parameters
        ----------
        e : Exception
            The error raised.
        """
        raise e

    async def on_send(self, message):
        """
        Called every time a message is sent to the server.

        Parameters
        ----------
        message
            The message sent to the server.
        """
        pass



class AudioPlayer:
    def __init__(self, sampling_rate: int = 22050):
        self.sampling_rate = sampling_rate
        self.audio_player = None
        self.stream = None

    def open(self):
        try:
            import pyaudio
        except ModuleNotFoundError:
            message = '`pip install pyaudio` required to use `AudioPlayer`'
            raise ModuleNotFoundError(message)

        self.audio_player = pyaudio.PyAudio()  # create the PyAudio player

        # start the audio stream, which will play audio as and when required
        self.stream = self.audio_player.open(
            format=pyaudio.paInt16, channels=1, rate=self.sampling_rate, output=True
        )

    def play(self, audio_bytes: bytes):
        if self.stream:
            self.stream.write(audio_bytes)

    def close(self):
        if self.stream:
            self.stream.stop_stream()
            self.stream.close()
            self.stream = None
        if self.audio_player:
            self.audio_player.terminate()
            self.audio_player = None

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()



"""
This module contains the `PyAudio` specific callbacks for the websocket client.

These are automatically used when the `play_audio` parameter on the `NeuphonicWebsocketClient` is set to `True` which
is the default value.
These functions are called on open (`setup_pyaudio`), on message (`play_audio`) and on close (`teardown_pyaudio`) of the
websocket client.
PyAudio must be installed to use these functions, otherwise if `play_audio` is set to `True` and PyAudio is not installed
then a warning will be raised.
"""

from base64 import b64decode


async def setup_pyaudio(self, sampling_rate=22050) -> None:
    """
    Create PyAudio resources needed to play audio, when the websocket opens.

    Parameters
    ----------
    self
        A NeuphonicWebsocketClient instance.
    """
    try:
        import pyaudio
    except ModuleNotFoundError:
        message = (
            '`pip install pyaudio` required to play audio using functions in '
            '`pyneuphonic.websocket.common.pyaudio`.'
        )
        raise ModuleNotFoundError(message)

    self.audio_player = pyaudio.PyAudio()  # create the PyAudio player

    # start the audio stream, which will play audio as and when required
    self.stream = self.audio_player.open(
        format=pyaudio.paInt16, channels=1, rate=sampling_rate, output=True
    )


async def play_audio(self, message: dict):
    """
    Callback to handle incoming audio messages.

    Appends audio byte data to the open PyAudio stream to play the audio in real-time.

    Parameters
    ----------
    self
        A NeuphonicWebsocketClient instance.
    message
        The message received from the websocket, as a dict object.
    """
    audio_bytes = b64decode(message['data']['audio'])
    self.stream.write(audio_bytes)  # type: ignore[attr-defined]


async def teardown_pyaudio(self):
    """
    Close the PyAudio resources opened up by on_open.

    This function will stop the audio stream and close the PyAudio player, freeing up the resources used by PyAudio.

    Parameters
    ----------
    self
        A NeuphonicWebsocketClient instance.
    """
    self.stream.stop_stream()  # type: ignore[attr-defined]
    self.stream.close()  # type: ignore[attr-defined]
    self.audio_player.terminate()  # type: ignore[attr-defined]
    self._logger.debug('Terminated PyAudio resources.')



import asyncio
from typing import Callable, Awaitable


class SubscriptableAsyncByteArray(bytearray):
    def __init__(self):
        """
        A derived class on the built-in Python `bytearray` object, providing subscription capabilities.

        **Why is this class useful?**
        Incoming audio can be stored in this object so that one or multiple subscribers can be notified of the new audio
        data, and process it as required. This class is used by the pyaudio and sounddevice examples in
        `pyneuphonic.websocket.common` for storing incoming audio data and playing it in real-time.

        The class is provided for usage however required, or can be extended as needed.
        """
        super().__init__()
        self.subscribers = []
        self.notify_lock = asyncio.Lock()

    def subscribe(self, callback: Callable[[bytes], Awaitable]):
        """
        Subscribe to any updates with a callback function.

        Parameters
        ----------
        callback
            The callback function. Takes the newly appended bytes and processes them. Must be an async function taking
            a single argument (the new bytes).
        """
        self.subscribers.append(callback)

    async def extend(self, new_bytes: bytes):
        """
        Extend the bytearray with new bytes and notify subscribers.

        Subscribers are called by passing in the new bytes as an argument.

        Parameters
        ----------
        new_bytes : bytes
            The new bytes to extend the bytearray with.

        """
        super().extend(new_bytes)

        # lock this next part of code so that only 1 coroutine enters this block at a time
        async with self.notify_lock:
            await self._notify_subscribers(new_bytes)

    async def _notify_subscribers(self, new_bytes):
        await asyncio.gather(*(callback(new_bytes) for callback in self.subscribers))



"""This websocket implementation has been DEPRECATED"""

from .libs import setup_pyaudio, teardown_pyaudio, play_audio, send_async_generator
from .neuphonic_websocket_client import NeuphonicWebsocketClient



from typing import List

import httpx

from ._endpoint import Endpoint
from .models import VoicesResponse, VoiceItem


class Voices(Endpoint):
    def get(self) -> List[VoiceItem]:
        """List all the voices."""
        response = httpx.get(
            f'{self.http_url}/voices',
            headers=self.headers,
            timeout=self.timeout,
        )

        if not response.is_success:
            raise httpx.HTTPStatusError(
                f'Failed to fetch voices. Status code: {response.status_code}. Error: {response.text}',
                request=response.request,
                response=response,
            )

        voice_response = VoicesResponse(**response.json())

        return voice_response.data.voices



import httpx
import json
from typing import Generator, AsyncGenerator, Optional, Union

from pyneuphonic._endpoint import Endpoint
from pyneuphonic.models import TTSConfig, SSEResponse, to_dict


class SSEClientBase(Endpoint):
    """Contains shared functions used by both the SSEClient and the AsyncSSE Client."""

    def _parse_message(self, message: str) -> Optional[SSEResponse]:
        """Parse each response from the server. Return as an SSEResponse object."""
        lines = message.split('\n')
        data = {}

        for line in lines:
            if not line.strip():
                continue

            try:
                key, value = line.split(': ', 1)
            except Exception as e:
                raise ValueError(f'Client request was malformed or incorrect: {line}')

            if key == 'data':
                data[key] = value

        if 'data' in data:
            return SSEResponse(**json.loads(data['data']))

        return None


class SSEClient(SSEClientBase):
    def jwt_auth(self) -> None:
        """
        Authenticate with the server to obtain a JWT token.

        This method sends a POST request to the server's authentication endpoint to exchange the
        API key for a JWT token.
        The token is then added to the headers for subsequent requests.
        Using JWT auth for subsequent requests speeds up the time to first audio byte when using
        the SSEClient.send method.
        Using JWT auth is recommended for situations where latency is a priority.

        Raises
        ------
        httpx.HTTPStatusError
            If the authentication request fails, an HTTPStatusError is raised with
            details about the failure.
        """
        response = httpx.post(
            f'{self.http_url}/sse/auth', headers=self.headers, timeout=self.timeout
        )

        if not response.is_success:
            raise httpx.HTTPStatusError(
                f'Failed to authenticate for a JWT. Status code: {response.status_code}. Error: {response.text}',
                request=response.request,
                response=response,
            )

        jwt_token = response.json()['data']['jwt_token']

        self.headers['Authorization'] = f'Bearer: {jwt_token}'

    def send(
        self, text: str, tts_config: Union[TTSConfig, dict] = TTSConfig()
    ) -> Generator[SSEResponse, None, None]:
        """
        Send a text to the TTS (text-to-speech) service and receive a stream of SSEResponse messages.

        Parameters
        ----------
        text : str
            The text to be converted to speech.
        tts_config : Union[TTSConfig, dict], optional
            The TTS configuration settings. Can be an instance of TTSConfig or a dictionary which
            will be parsed into a TTSConfig.

        Yields
        ------
        Generator[SSEResponse, None, None]
            A generator yielding SSEResponse messages.
        """
        if not isinstance(tts_config, TTSConfig):
            tts_config = TTSConfig(**tts_config)

        assert isinstance(text, str), '`text` should be an instance of type `str`.'

        with httpx.stream(
            method='POST',
            url=f'{self.http_url}/sse/speak/{tts_config.language_id}',
            headers=self.headers,
            json={'text': text, 'model': to_dict(tts_config)},
        ) as response:
            for message in response.iter_lines():
                parsed_message = self._parse_message(message)

                if parsed_message is not None:
                    yield parsed_message


class AsyncSSEClient(SSEClientBase):
    async def jwt_auth(self) -> None:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                f'{self.http_url}/sse/auth', headers=self.headers, timeout=self.timeout
            )

            if not response.is_success:
                raise httpx.HTTPStatusError(
                    f'Failed to authenticate for a JWT. Status code: {response.status_code}. Error: {response.text}',
                    request=response.request,
                    response=response,
                )

            jwt_token = response.json()['data']['jwt_token']
            self.headers['Authorization'] = f'Bearer: {jwt_token}'

    async def send(
        self, text: str, tts_config: Union[TTSConfig, dict] = TTSConfig()
    ) -> AsyncGenerator[SSEResponse, None]:
        if not isinstance(tts_config, TTSConfig):
            tts_config = TTSConfig(**tts_config)

        assert isinstance(text, str), '`text` should be an instance of type `str`.'

        async with httpx.AsyncClient() as client:
            async with client.stream(
                method='POST',
                url=f'{self.http_url}/sse/speak/{tts_config.language_id}',
                headers=self.headers,
                json={'text': text, 'model': to_dict(tts_config)},
            ) as response:
                async for message in response.aiter_lines():
                    parsed_message = self._parse_message(message)

                    if parsed_message is not None:
                        yield parsed_message



import ssl
import certifi


class Endpoint:
    def __init__(
        self,
        api_key: str,
        base_url: str,
        timeout: int = 10,
    ):
        self._api_key = api_key
        self._base_url = base_url
        self.timeout = timeout

        self.headers = {
            'x-api-key': self._api_key,
        }

    @property
    def base_url(self):
        return self._base_url

    def _is_localhost(self):
        return True if 'localhost' in self.base_url else False

    @property
    def http_url(self):
        prefix = 'http' if self._is_localhost() else 'https'
        return f'{prefix}://{self.base_url}'

    @property
    def ws_url(self):
        prefix = 'ws' if self._is_localhost() else 'wss'
        return f'{prefix}://{self.base_url}'

    @property
    def ssl_context(self):
        ssl_context = (
            None
            if self._is_localhost()
            else ssl.create_default_context(cafile=certifi.where())
        )

        return ssl_context
